import re

async def if_(value, invert=False, _parser_context=None):
    ''' Standard if/else statement.

    Parameters
    ----------
    value : bool
        The value to check. If `True` then the first block will be executed, otherwise the second block
        (the one after the `{{else}}`) will be executed.
    invert : bool
        If `True` then the value will be inverted before checking.
    '''
    block_content = _parser_context['block_content']
    assert len(block_content) in [1,3] # we don't support elseif yet...
    options = [block_content[0]]
    for i in range(1, len(block_content), 2):
        assert re.match(r"{{~?else~?}}",  block_content[i].text), "Expected else statement"
        options.append(block_content[i+1])

    # if isinstance(value, str):
    #     value2 = value
    #     value = value.lower().strip() in ["true", "yes", "on", "t", "y", "ok", "okay"]
    
    if invert:
        value = not value
    
    if value:
        return await _parser_context['parser'].visit(options[0])
    elif len(options) > 1:
        return await _parser_context['parser'].visit(options[1])
    else:
        return ""
if_.is_block = True


import ast
import inspect
import re
import html
import uuid
import sys
import parsimonious
import logging
import copy
import asyncio
import pathlib
import os
import traceback
import time
import datetime
from .llms import _openai
from . import _utils
from ._program_executor import ProgramExecutor
from . import library
import guidance
log = logging.getLogger(__name__)

# load the javascript client code
file_path = pathlib.Path(__file__).parent.parent.absolute()
with open(file_path / "guidance" / "resources" / "main.js", encoding="utf-8") as f:
    js_data = f.read()

class Program:
    ''' A program template that can be compiled and executed to generate a new filled in (executed) program.

    Note that as the template gets executed {{!-- handlebars comment markers --}} get left in
    the generated output to mark where template tags used to be.
    '''

    def __init__(self, text, llm=None, cache_seed=0, logprobs=None, silent='auto', async_mode=False, stream='auto', caching=None, await_missing=False, **kwargs):
        """ Create a new Program object from a program string.

        Parameters
        ----------
        text : str
            The program string to use as a guidance template.
        llm : guidance.llms.LLM (defaults to guidance.llm)
            The language model to use for executing the program.
        cache_seed : int (default 0) or None
            The seed to use for the cache. If you want to use the same cache for multiple programs
            you can set this to the same value for all of them. Set this to None to disable caching.
            Caching is enabled by default, and saves calls that have tempurature=0, and also saves
            higher temperature calls but uses different seed for each call.
        logprobs : int or None (default)
            The number of logprobs to return from the language model for each token. (not well supported yet,
            since some endpoints don't support it)
        silent : bool (default False)
            If True, the program will not display any output. This is useful for programs that are
            only used to generate variables for other programs.
        async_mode : bool (default False)
            If True, the program will be executed asynchronously. This is useful for programs that
            take a long time to run, or that need to be run in parallel.
        stream : bool (default False)
            If True, the program will try to stream all the results from the LLM token by token.
        caching : bool (default None)
            If True, the program will cache the results of the LLM. If False, it will not cache the results.
            If None, it will use the default caching setting from the LLM.
        await_missing : bool (default False)
            If True, the program will automatically await any missing variables. This means the program
            will stop executation at that point and return a paritally executed program. This is useful
            for executing programs on different machines, for example shipping a program to a GPU machine
            then waiting for the results to come back for any local processing, then shipping it back to
            the GPU machine to continue execution.
        """

        # see if we were given a raw function instead of a string template
        # if so, convert it to a string template that calls the function
        if not isinstance(text, str):
            if callable(text):
                sig = inspect.signature(text)
                args = ""
                for name,_ in sig.parameters.items():
                    args += f" {name}={name}"
                fname = _utils.find_func_name(text, kwargs)
                kwargs[fname] = text
                text = "{{set (%s%s)}}" % (fname, args)
        
        # save the given parameters
        self._text = text
        self.llm = llm or getattr(guidance, "llm", None)
        self.cache_seed = cache_seed
        self.caching = caching
        self.logprobs = logprobs
        self.async_mode = async_mode
        self.silent = silent
        self.stream = stream
        self.await_missing = await_missing
        if self.silent == "auto":
            self.silent = not _utils.is_interactive()
        
        # set our variables
        self._variables = {}
        self._variables.update(_built_ins)
        self._variables.update({
            "llm": llm
        })
        self._variables.update(kwargs)
        
        # set internal state variables
        self._id = str(uuid.uuid4())
        
        self._comm = None # front end communication object
        self._executor = None # the ProgramExecutor object that is running the program
        self._last_display_update = 0 # the last time we updated the display (used for throttling updates)
        self._execute_complete = asyncio.Event() # fires when the program is done executing to resolve __await__
        self._displaying = not self.silent # if we are displaying we need to update the display as we execute
        self._displayed = False # marks if we have been displayed in the client yet
        self._displaying_html = False # if we are displaying html (vs. text)

        # throttle the display updates
        if os.environ.get("VSCODE_CWD", None) is not None:
            self.display_throttle_limit = 0.1 # VSCode has a bug that causes flashing, so we slow down the display
        else:
            self.display_throttle_limit = 0.1 # the minimum time between display updates
        self.update_display = DisplayThrottler(self._update_display, self.display_throttle_limit)

        # see if we are in an ipython environment
        try:
            self._ipython = get_ipython()
        except:
            self._ipython = None
        
        # if we are echoing in ipython we assume we can display html
        if self._ipython and not self.silent:
            self._displaying_html = True
    
    def __repr__(self):
        return self.text
    
    def __getitem__(self, key):
        return self._variables[key]
    
    def _interface_event(self, msg):
        """ Handle an event from the front end.
        """
        if msg["event"] == "stop":
            self._executor.stop()
        elif msg["event"] == "opened":
            pass # we don't need to do anything here because the first time we display we'll send the html
        pass

    def _ipython_display_(self):
        """ Display the program in the ipython notebook.
        """

        log.debug(f"displaying program in _ipython_display_ with self._comm={self._comm}, self.id={self._id}")
        
        # mark that we are displaying (and so future execution updates should be displayed)
        self._displaying = True
        self._displaying_html = True
        
        # build and display the html
        html = self._build_html(self.marked_text)
        self._display_html(html)
        

    async def _await_finish_execute(self):
        """ Used by self.__await__ to wait for the program to complete.
        """
        await self._execute_complete.wait() # wait for the program to finish executing
        return self

    def __await__(self):
        return self._await_finish_execute().__await__()
        
    def __call__(self, **kwargs):
        """ Execute this program with the given variable values and return a new executed/executing program.

        Note that the returned program might not be fully executed if `stream=True`. When streaming you need to
        use the python `await` keyword if you want to ensure the program is finished (note that is different than
        the `await` guidance langauge command, which will cause the program to stop execution at that point).
        """

        # merge the given kwargs with the current variables
        kwargs = {**{
            "async_mode": self.async_mode,
            "stream": self.stream,
            "silent": self.silent,
            "cache_seed": self.cache_seed,
            "caching": self.caching,
            "logprobs": self.logprobs,
            "await_missing": self.await_missing,
            "llm": self.llm,
        }, **kwargs}

        log.debug(f"in __call__ with kwargs: {kwargs}")

        # create a new program object that we will execute in-place
        new_program = Program(
            text=self.marked_text,

            # copy the (non-function) variables so that we don't modify the original program during execution
            # TODO: what about functions? should we copy them too?
            **{**{k: v if callable(v) else copy.deepcopy(v) for k,v in self._variables.items()}, **kwargs}
        )

        # create an executor for the new program (this also marks the program as executing)
        new_program._executor = ProgramExecutor(new_program)
        
        # if we are in async mode schedule the program in the current event loop
        if new_program.async_mode:
            loop = asyncio.get_event_loop()
            assert loop.is_running(), "The program is in async mode but there is no asyncio event loop running! Start one and try again."
            loop.create_task(new_program.update_display.run()) # start the display updater
            loop.create_task(new_program.execute())

        # if we are not in async mode, we need to create a new event loop and run the program in it until it is done
        else:
            loop = asyncio.new_event_loop()
            loop.create_task(new_program.update_display.run()) # start the display updater
            loop.run_until_complete(new_program.execute())
    
        return new_program

    def _update_display(self, last=False):
        """ Updates the display with the current marked text after debouncing.

        Parameters
        ----------
        last : bool
            If True, this is the last update and we should clear the send queue and prepare the
            UI for saving etc.
        force : bool
            If True, we will update the display even if it would otherwise be throttled.
        """

        log.debug(f"Updating display (last={last}, self._displaying={self._displaying}, self._comm={self._comm})")

        # this is always called during execution, and we only want to update the display if we are displaying
        if not self._displaying:
            return
        
        # debounce the display updates
        # now = time.time()
        # log.debug(now - self._last_display_update)
        # debounce_delay = self.display_throttle_limit if self._comm and self._comm.is_open else self.display_throttle_limit_low
        # if last or (now - self._last_display_update > debounce_delay):
        if self._displaying_html:
            out = self._build_html(self.marked_text)
            
            # clear the send queue if this is the last update
            if last and self._comm:
                self._comm.clear_send_queue()
            
            # send an update to the front end client if we have one...
            # TODO: we would like to call `display` for the last update so NB saving works, but see https://github.com/microsoft/vscode-jupyter/issues/13243 
            if self._displayed and self._comm and self._comm.is_open: #(not last or self._comm.is_open):
                log.debug(f"Updating display send message to front end")
                # log.debug(out)
                self._comm.send({"replace": out})
                if last:
                    self._comm.send({"event": "complete"})
            
            # ...otherwise dump the client to the font end
            else:
                log.debug(f"Updating display dump to front end")
                from IPython.display import clear_output, display
                if self._displayed:
                    clear_output(wait=True) # TODO: should use wait=True but that doesn't work in VSCode until after the April 2023 release

                self._display_html(out)
        
        self._last_display_update = time.time()

    def _display_html(self, html):
        from IPython.display import display

        # create the comm object if we don't have one
        if self._comm is None:
            self._comm = _utils.JupyterComm(self._id, self._ipython, self._interface_event)
        
        # dump the html to the front end
        html = f"""<div id="guidance-stop-button-{self._id}" style="cursor: pointer; margin: 0px; display: none; float: right; padding: 3px; border-radius: 4px 4px 4px 4px; border: 0px solid rgba(127, 127, 127, 1); padding-left: 10px; padding-right: 10px; font-size: 13px; background-color: rgba(127, 127, 127, 0.25);">Stop program</div><div id="guidance-content-{self._id}">{html}</div>
<script type="text/javascript">{js_data}; window._guidanceDisplay("{self._id}");</script>"""
        display({"text/html": html}, display_id=self._id, raw=True, clear=True, include=["text/html"])
        self._displayed = True

    async def execute(self):
        """ Execute the current program.

        Note that as execution progresses the program will be incrementally converted
        from a template into a completed string (with variables stored). At each point
        in this process the current template remains valid.
        """

        log.debug(f"Executing program (self.async_mode={self.async_mode}, self.silent={self.silent}, self._displaying_html={self._displaying_html})")
        
        # if we are already displaying html, we need to yeild to the event loop so the jupyter comm can initialize
        if self._displaying_html:
            await asyncio.sleep(0)
        
        # run the program and capture the output
        if self.llm is None:
            await self._executor.run(None)
        else:
            with self.llm.session(asynchronous=True) as llm_session:
                await self._executor.run(llm_session)
        self._text = self._executor.prefix

        # delete the executor and so mark the program as not executing
        self._executor = None

        # update the display with the final output
        self.update_display(last=True)
        await self.update_display.done()

        # fire an event noting that execution is complete (this will release any await calls waiting on the program)
        self._execute_complete.set()
    
    def __getitem__(self, key):
        return self._variables[key]
    
    def __contains__(self, key):
        return key in self._variables
    
    def __delitem__(self, key):
        del self._variables[key]
    
    def variables(self, built_ins=False):
        """ Returns a dictionary of the variables in the program.

        Parameters
        ----------
        built_ins : bool
            If True, built-in variables will be included in the returned dictionary.
        """
        return {k: v for k,v in self._variables.items() if built_ins or not (k in _built_ins and callable(_built_ins[k]))}
    
    @property
    def text(self):
        # strip out the markers for the unformatted output
        return _utils.strip_markers(self.marked_text)
    
    @property
    def marked_text(self):
        if self._executor is not None:
            return self._executor.prefix
        else:
            return self._text
    
    def _build_html(self, text, last=False):
        output = text

        def undo_html_encode(x):
            return x.replace("&amp;#123;", "{").replace("&amp;#125;", "}").replace("&amp;#36;", "$")

        def start_generate_or_select(x):
            no_echo = "echo=False" in x.group(1)
            alpha = 1.0 if no_echo else 1.0
            
            # script that toggles the viisibility of the next element
            click_script = 'var e = this.nextElementSibling; if (e.style.display == "inline") { e.style.display = "none"; this.style.borderRight = "1px solid rgba(0, 165, 0, 0.25)"; } else { e.style.display = "inline"; this.style.borderRight = "0px";}'

            if no_echo:
                out = f'''<div style='background-color: rgba(0, 165, 0, 0.25); border-radius: 4px 0px 0px 4px; border: 1px solid rgba(0, 165, 0, 1); padding-left: 3px; padding-right: 3px; user-select: none; color: rgb(0, 165, 0, 1.0); display: inline; font-weight: normal; cursor: pointer' onClick='{click_script}'>no echo</div>'''
                out += "<span style='background-color: rgba(0, 165, 0, 0.25); opacity: {}; display: none;' title='{}'>".format(alpha, undo_html_encode(x.group(1)))
            else:
                out = "<span style='background-color: rgba(0, 165, 0, 0.25); opacity: {}; display: inline;' title='{}'>".format(alpha, undo_html_encode(x.group(1)))
            return out
        
        def start_each(x):
            no_echo = "echo=False" in x.group(1)
            alpha = 0.5 if no_echo else 1.0
            color = "rgba(165, 165, 165, 0.1)" #if "geneach" not in x.group(1) else "rgba(0, 165, 0, 0.1)"
            return "<span style='opacity: {}; display: inline; background-color: {};' title='{}'>".format(alpha, color, undo_html_encode(x.group(1)))
        
        def start_block(x):
            escaped_tag = undo_html_encode(x.group(1))
            if "hidden=True" in escaped_tag:
                display = "inline" # none (we actively stip hidden tags right now so we don't need this until we support the UX to show hidden stuff)
            else:
                display = "inline"
            return f"<span style='background-color: rgba(165, 165, 165, 0.1); display: {display};' title='{escaped_tag}'>"
        
        def role_box(x):
            # name = x.group(3).lower() # standardize to lowercase for display
            # content = x.group(4)

            content = x.group(3)
            tag_text = undo_html_encode(x.group(2))
            role_name = x.group(1)
            
            # if we have a generic role tag then the role name is an attribute
            if role_name == "role":
                role_name = re.search(r"name=([^ ]*)", tag_text).group(1)
            
            start_pattern = html.escape(self.llm.role_start(role_name)).replace("|", r"\|")
            end_pattern = html.escape(self.llm.role_end(role_name)).replace("|", r"\|")

            # strip the start and end patterns from the content
            content = re.sub("^" + start_pattern, "", content, flags=re.DOTALL)
            content = re.sub(end_pattern + "$", "", content, flags=re.DOTALL)

            
            # one div that contains two divs, where the left of the two inner divs has a fixed width of 100px
            # """<div style='display: flex;'>
            #     <div style='width: 100px; border-right: 1px solid rgba(127, 127, 127, 0.2); padding-right: 5px; margin-right: 5px;'>{name}</div>
            #     <div>{content}</div>
            # </div>"""

            # return f'''<div style="border-left: 1px solid rgba(127, 127, 127, 0.2); margin-top: 10px; padding-left: 5px;"><span style="color: rgba(127,127,127,0.5)">{name}</span>
# {content}</div>'''

            return f"<div style='display: flex; border-bottom: 1px solid rgba(127, 127, 127, 0.2); align-items: center;'><div style='flex: 0 0 80px; opacity: 0.5;'>{role_name.lower()}</div><div style='flex-grow: 1; padding: 5px; padding-top: 10px; padding-bottom: 10px; margin-top: 0px; white-space: pre-wrap; margin-bottom: 0px;'>{content}</div></div>"

        display_out = html.escape(output)
        # log.debug(display_out)

        
        # start_pattern = html.escape(self.llm.role_start("(.*?)")).replace("|", r"\|")
        # end_pattern = html.escape(self.llm.role_end("(.*?)")).replace("|", r"\|")
        # display_out = re.sub(r"[\s]+({{!--.*?--}})?"+start_pattern, r"\1"+start_pattern.replace("(.*?)", r"\1").replace(r"\|", "|"), display_out, flags=re.DOTALL)
        # display_out = re.sub(start_pattern + "(.*?)" + end_pattern, role_box, display_out, flags=re.DOTALL)
        # log.debug(display_out)

        # strip out hidden blocks (might want to make a better UI for this at some point)
        display_out = re.sub(r"{{!--GMARKER_START[^}]*--}}{{!--GHIDDEN:(.*?)--}}{{!--GMARKER_END[^}]*--}}", "", display_out, flags=re.DOTALL)
        
        # if we have role markers, we wrap them in special formatting
        if re.search(r"{{!--GMARKER_START_(role|system|user|assistant)", display_out) is not None:

            # start_pattern = html.escape(self.llm.role_start("assistant")).replace("|", r"\|").replace(r"assistant", r"([^\n]*)").replace(r"ASSISTANT", r"([^\n]*)")
            # end_pattern = html.escape(self.llm.role_end("assistant")).replace("|", r"\|").replace(r"assistant", r"([^\n]*)").replace(r"ASSISTANT", r"([^\n]*)")
            
            # strip whitespace before role markers
            display_out = re.sub(r"\s*{{!--GMARKER_START_(role|system|user|assistant)\$(.*?)--}}", r"{{!--GMARKER_START_\1$\2--}}", display_out, flags=re.DOTALL)

            # strip whitespace after role markers
            # TODO: support end_patterns with capture groups
            display_out = re.sub(r"{{!--GMARKER_END_(role|system|user|assistant)\$(.*?)--}}\s*", r"{{!--GMARKER_END_\1$\2--}}", display_out, flags=re.DOTALL)

            # wrap role markers in nice formatting
            display_out = re.sub(r"{{!--GMARKER_START_(role|system|user|assistant)\$(.*?)--}}" + "(.*?)" + r"{{!--GMARKER_END_(role|system|user|assistant)\$(.*?)--}}", role_box, display_out, flags=re.DOTALL)

            # wrap unfinished role markers in nice formatting
            display_out = re.sub(r"{{!--GMARKER_START_(role|system|user|assistant)\$(.*?)--}}" + "(.*)", role_box, display_out, flags=re.DOTALL)
        
        display_out = re.sub(r"(\{\{generate.*?\}\})", r"<span style='background-color: rgba(0, 165, 0, 0.25);'>\1</span>", display_out, flags=re.DOTALL)
        display_out = re.sub(r"(\{\{#select\{\{/select.*?\}\})", r"<span style='background-color: rgba(0, 165, 0, 0.25);'>\1</span>", display_out, flags=re.DOTALL)
        display_out = re.sub(r"(\{\{#each [^'\"].*?\{\{/each.*?\}\})", r"<span style='background-color: rgba(0, 138.56128016, 250.76166089, 0.25);'>\1</span>", display_out, flags=re.DOTALL)
        display_out = re.sub(r"(\{\{(?!\!)(?!generate)(?!#select)(?!#each)(?!/each)(?!/select).*?\}\})", r"<span style='background-color: rgba(0, 138.56128016, 250.76166089, 0.25);'>\1</span>", display_out, flags=re.DOTALL)
                

        # format the generate command results
        display_out = re.sub(r"{{!--GMARKER_START_gen\$([^\$]*)\$--}}", start_generate_or_select, display_out)
        display_out = display_out.replace("{{!--GMARKER_END_gen$$--}}", "</span>")
        def click_loop_start(id, total_count, echo, color):
            click_script = '''
function cycle_IDVAL(button_el) {
    var i = 0;
    while (i < 50) {
        var el = document.getElementById("IDVAL_" + i);
        if (el.style.display == "inline") {
            el.style.display = "none";
            var next_el = document.getElementById("IDVAL_" + (i+1));
            if (!next_el) {
                next_el = document.getElementById("IDVAL_0");
            }
            if (next_el) {
                next_el.style.display = "inline";
            }
            break;
        }
        i += 1;
    }
    button_el.innerHTML = (((i+1) % TOTALCOUNT) + 1)  + "/" + TOTALCOUNT;
}
cycle_IDVAL(this);'''.replace("IDVAL", id).replace("TOTALCOUNT", str(total_count)).replace("\n", "")
            out = f'''<div style='background: rgba(255, 255, 255, 0.0); border-radius: 4px 0px 0px 4px; border: 1px solid {color}; border-right: 0px; padding-left: 3px; padding-right: 3px; user-select: none; color: {color}; display: inline; font-weight: normal; cursor: pointer' onClick='{click_script}'>1/{total_count}</div>'''
            out += f"<div style='display: inline;' id='{id}_0'>"
            return out
        def click_loop_mid(id, index, echo):
            alpha = 1.0 if not echo else 0.5
            out = f"</div><div style='display: none; opacity: {alpha}' id='{id}_{index}'>"
            return out
        display_out = re.sub(
            r"{{!--GMARKERmany_generate_start_([^_]+)_([0-9]+)\$([^\$]*)\$--}}",
            lambda x: click_loop_start(x.group(3), int(x.group(2)), x.group(1) == "True", "rgba(0, 165, 0, 0.25)"),
            display_out
        )
        display_out = re.sub(
            r"(?:--}})?{{!--GMARKERmany_generate_([^_]+)_([0-9]+)\$([^\$]*)\$--}}{{!--G ",
            lambda x: click_loop_mid(x.group(3), int(x.group(2)), x.group(1) == "True"),
            display_out
        )
        display_out = re.sub(r"--}}{{!--GMARKERmany_generate_end\$([^\$]*)\$--}}", "</div>", display_out)

        # format the each command results
        display_out = re.sub(r"{{!--GMARKER_START_each\$([^\$]*)\$--}}", start_each, display_out)
        display_out = re.sub(
            r"{{!--GMARKER_each_noecho_start_([^_]+)_([0-9]+)\$([^\$]*)\$--}}",
            lambda x: click_loop_start(x.group(3), int(x.group(2)), False, "rgb(100, 100, 100, 1)"),
            display_out
        )
        display_out = re.sub(
            r"{{!--GMARKER_each_noecho_([^_]+)_([0-9]+)\$([^\$]*)\$--}}",
            lambda x: click_loop_mid(x.group(3), int(x.group(2)), False),
            display_out
        )
        display_out = re.sub(r"{{!--GMARKER_each_noecho_end\$([^\$]*)\$--}}", "</div>", display_out)

        # format the geneach command results
        display_out = re.sub(r"{{!--GMARKER_START_geneach\$([^\$]*)\$--}}", start_each, display_out)
        
        # format the set command results
        display_out = re.sub(r"{{!--GMARKER_set\$([^\$]*)\$--}}", r"<div style='background-color: rgba(165, 165, 165, 0); border-radius: 4px 4px 4px 4px; border: 1px solid rgba(165, 165, 165, 1); border-left: 2px solid rgba(165, 165, 165, 1); border-right: 2px solid rgba(165, 165, 165, 1); padding-left: 0px; padding-right: 3px; color: rgb(165, 165, 165, 1.0); display: inline; font-weight: normal; overflow: hidden;'><div style='display: inline; background: rgba(165, 165, 165, 1); padding-right: 5px; padding-left: 4px; margin-right: 3px; color: #fff'>set</div>\1</div>", display_out)
        display_out = re.sub(r"{{!--GMARKER_START_set\$([^\$]*)\$--}}", lambda x: "<span style='display: inline;' title='{}'>".format(undo_html_encode(x.group(1))), display_out)

        display_out = re.sub(r"{{!--GMARKER_START_select\$([^\$]*)\$--}}", start_generate_or_select, display_out)
        display_out = display_out.replace("{{!--GMARKER_END_select$$--}}", "</span>")
        display_out = re.sub(r"{{!--GMARKER_START_variable_ref\$([^\$]*)\$--}}", lambda x: "<span style='background-color: rgba(0, 138.56128016, 250.76166089, 0.25); display: inline;' title='{}'>".format(undo_html_encode(x.group(1))), display_out)
        display_out = display_out.replace("{{!--GMARKER_END_variable_ref$$--}}", "</span>")
        display_out = display_out.replace("{{!--GMARKER_each$$--}}", "")#<div style='border-left: 1px dashed rgb(0, 0, 0, .2); border-top: 0px solid rgb(0, 0, 0, .2); margin-right: -4px; display: inline; width: 4px; height: 24px;'></div>")
        display_out = re.sub(r"{{!--GMARKER_START_block\$([^\$]*)\$--}}", start_block, display_out)
        display_out = re.sub(r"{{!--GMARKER_START_([^\$]*)\$([^\$]*)\$--}}", lambda x: "<span style='background-color: rgba(0, 138.56128016, 250.76166089, 0.25); display: inline;' title='{}'>".format(undo_html_encode(x.group(2))), display_out)
        display_out = re.sub(r"{{!--GMARKER_END_([^\$]*)\$\$--}}", "</span>", display_out)
        
        # display_out = re.sub(' and (?=.* and )', ', ', display_out)

        # strip out comments
        display_out = re.sub(r"{{~?!.*?}}", "", display_out)

        # re.sub(r"<div class='strip_leading_whitespace'")

        display_out = add_spaces(display_out)
        display_out = "<pre style='margin: 0px; padding: 0px; padding-left: 8px; margin-left: -8px; border-radius: 0px; border-left: 1px solid rgba(127, 127, 127, 0.2); white-space: pre-wrap; font-family: ColfaxAI, Arial; font-size: 15px; line-height: 23px;'>"+display_out+"</pre>"

        return display_out

def add_spaces(s):
    """ This adds spaces so the browser will show leading and trailing newlines.
    """
    if s.startswith("\n"):
        s = " " + s
    if s.endswith("\n"):
        s = s + " "
    return s

_built_ins = {
    "gen": library.gen,
    "each": library.each,
    "geneach": library.geneach,
    "select": library.select,
    "if": library.if_,
    "unless": library.unless,
    "add": library.add,
    "subtract": library.subtract,
    "strip": library.strip,
    "block": library.block,
    "set": library.set,
    "await": library.await_,
    "role": library.role,
    "user": library.user,
    "system": library.system,
    "assistant": library.assistant,
    "break": library.break_,
    "equal": library.equal,
}

class DisplayThrottler():
    def __init__(self, display_function, throttle_limit):
        self.display_function = display_function
        self.throttle_limit = throttle_limit
        self._done = False
        self.last_time = 0
    
    async def run(self):
        self._data_event = asyncio.Event()
        self._done_event = asyncio.Event()
        while True:
            await self._data_event.wait()
            now = time.time()
            log.info("in DisplayThrottler run loop -- now: {}, last_time: {}, throttle_limit: {}".format(now, self.last_time, self.throttle_limit))
            if self._done or now - self.last_time >= self.throttle_limit:
                try:
                    self.display_function(last=self._done)
                except Exception as e:
                    self._done = True
                    raise e
                finally:
                    self.last_time = now
                    self._data_event.clear()
                    if self._done:
                        self._done_event.set()
                        break
            else:
                await asyncio.sleep(self.throttle_limit - (now - self.last_time))

    def __call__(self, last=False):
        if last:
            self._done = True
        self._data_event.set()

    async def done(self):
        return await self._done_event.wait()


import random

class Random():
    def __init__(self, items, k=1):
        ''' Create a selector that chooses a random set of instances.

        Parameters
        ----------
        k : int
            The number of instances to choose.
        '''
        self.items = items
        self.k = k
    
    def __call__(self):
        ''' Selects a random set of instances.
        '''
        return random.choice(self.items, k=self.k)
    
class NGramOverlap():
    def __init__(self, items, k=1):
        ''' Create a selector that chooses a random set of instances.

        Parameters
        ----------
        k : int
            The number of instances to choose.
        '''
        # TODO: Implement this.
        # It would be nice just use let the LangChain option cover this, but they have some odd dependencies on an example template.
        raise NotImplementedError("NGramOverlap is not implemented yet.")
    
    def __call__(self):
        ''' Selects a random set of instances.
        '''
        return self.items

def _word_tokenizer(text):
    ''' Tokenizes a string by white space.
    '''
    return text.split()

class TokenLimit():
    
    def __init__(self, items, max_tokens=20):
        ''' Create a selector that limits the number of tokens in a list of items.

        Parameters
        ----------
        items : list
            A list of items to select from.
        max_tokens : int
            The maximum number of tokens to allow.
        '''

        self.items = items
        self.max_tokens = max_tokens
    
    def __call__(self, template_context=None):
        ''' Filters a list of items to a maximum number of tokens.

        Parameters
        ----------
        template_context : dict (optional)
            A dictionary of template context variables to use for token counting.
        
        Returns
        -------
        A list of items that fit within the token limit.
        '''
        
        if template_context is not None and "@tokenizer" in template_context:
            token_encoder = template_context["@tokenizer"].encode
        else:
            token_encoder = _word_tokenizer
        total_length = 0
        out = []
        for item in self.items:
            if template_context is not None and "@block_text" in template_context:
                context_new = template_context["@block_text"]
                if isinstance(item, dict):
                    for k in item:
                        context_new = context_new.replace("{{this."+k+"}}", item[k])
                else:
                    context_new = context_new.replace("{{this}}", item)
            else:
                context_new = " ".join([item[k] for k in item])
            new_length = len(token_encoder(context_new))
            if total_length + new_length <= self.max_tokens:
                total_length += new_length
                out.append(item)
        return out
    

class LangChain():
    def __init__(self, selector):
        ''' Create a selector from a LangChain ExampleSelector object.

        Parameters
        ----------
        selector : ExampleSelector
            A LangChain ExampleSelector object.
        
        Returns
        -------
        A selector that selects examples using a LangChain ExampleSelector object.
        '''
        self.selector = selector
    
    def __call__(self, **kwargs):
        ''' Select examples using a LangChain ExampleSelector object.

        Note that we use keyword arguments here instead of a single dictionary.
        '''
        out = self.selector.select_examples(kwargs)
        return out


from ._openai import OpenAI
from ._transformers import Transformers
from ._mock import Mock
from ._llm import LLM
from ._azure_openai import AzureOpenAI
from ._deep_speed import DeepSpeed
from . import transformers


import openai
import os
import time
import requests
import copy
import time
import asyncio
import types
import collections
import json
import re
from ._llm import LLM, LLMSession, SyncSession


class MalformedPromptException(Exception):
    pass
def prompt_to_messages(prompt):
    messages = []
    start_tags = re.findall(r'<\|im_start\|>', prompt)
    end_tags = re.findall(r'<\|im_end\|>', prompt)
    # if len(start_tags) != len(end_tags):
    #     raise MalformedPromptException("Malformed prompt: start and end tags are not properly paired")

    assert prompt.endswith("<|im_start|>assistant\n"), "When calling OpenAI chat models you must generate only directly inside the assistant role! The OpenAI API does not currently support partial assistant prompting."

    pattern = r'<\|im_start\|>(\w+)(.*?)(?=<\|im_end\|>|$)'
    matches = re.findall(pattern, prompt, re.DOTALL)

    if not matches:
        return [{'role': 'user', 'content': prompt.strip()}]

    for match in matches:
        role, content = match
        content = content.strip() # should we do this?
        messages.append({'role': role, 'content': content})

    return messages

def add_text_to_chat_mode_generator(chat_mode):
    for resp in chat_mode:
        if "choices" in resp:
            for c in resp['choices']:
                if "content" in c['delta']:
                    c['text'] = c['delta']['content']
                else:
                    break # the role markers are outside the generation in chat mode right now TODO: consider how this changes for uncontrained generation
            else:
                yield resp
        else:
            yield resp

def add_text_to_chat_mode(chat_mode):
    if isinstance(chat_mode, types.GeneratorType):
        return add_text_to_chat_mode_generator(chat_mode)
    else:
        for c in chat_mode['choices']:
            c['text'] = c['message']['content']
        return chat_mode
        

        # c['text'] = f'<|im_start|>{c["message"]["role"]}\n{c["message"]["content"]}<|im_end|>'

# model that need to use the chat completion API
chat_models = [
    "gpt-4",
    "gpt-3.5-turbo",
    "gpt-4-0314",
    "gpt-3.5-turbo-0301"
]

class OpenAI(LLM):
    cache = LLM._open_cache("_openai.diskcache")

    def __init__(self, model=None, caching=True, max_retries=5, max_calls_per_min=60, token=None, endpoint=None, temperature=0.0, chat_mode="auto"):
        super().__init__()

        # fill in default model value
        if model is None:
            model = os.environ.get("OPENAI_MODEL", None)
        if model is None:
            try:
                with open(os.path.expanduser('~/.openai_model'), 'r') as file:
                    model = file.read().replace('\n', '')
            except:
                pass

        # auto detect chat completion mode
        if chat_mode == "auto":
            if model in chat_models:
                chat_mode = True
            else:
                chat_mode = False
        
        # fill in default API key value
        if token is None: # get from environment variable
            token = os.environ.get("OPENAI_API_KEY", getattr(openai, "api_key", None))
        if token is not None and not token.startswith("sk-") and os.path.exists(os.path.expanduser(token)): # get from file
            with open(os.path.expanduser(token), 'r') as file:
                token = file.read().replace('\n', '')
        if token is None: # get from default file location
            try:
                with open(os.path.expanduser('~/.openai_api_key'), 'r') as file:
                    token = file.read().replace('\n', '')
            except:
                pass
        
        # fill in default endpoint value
        if endpoint is None:
            endpoint = os.environ.get("OPENAI_ENDPOINT", None)

        import tiktoken
        self._tokenizer = tiktoken.get_encoding(tiktoken.encoding_for_model(model).name)
        self.chat_mode = chat_mode
        
        self.model_name = model
        self.caching = caching
        self.max_retries = max_retries
        self.max_calls_per_min = max_calls_per_min
        if isinstance(token, str):
            token = token.replace("Bearer ", "")
        self.token = token
        self.endpoint = endpoint
        self.current_time = time.time()
        self.call_history = collections.deque()
        self.temperature = temperature

        if self.endpoint is None:
            self.caller = self._library_call
        else:
            self.caller = self._rest_call
            self._rest_headers = {
                "Content-Type": "application/json"
            }

    def session(self, asynchronous=False):
        if asynchronous:
            return OpenAISession(self)
        else:
            return SyncSession(OpenAISession(self))

    def role_start(self, role):
        assert self.chat_mode, "role_start() can only be used in chat mode"
        return "<|im_start|>"+role+"\n"
    
    def role_end(self, role=None):
        assert self.chat_mode, "role_end() can only be used in chat mode"
        return "<|im_end|>"
    
    @classmethod
    def stream_then_save(cls, gen, key):
        list_out = []
        for out in gen:
            list_out.append(out)
            yield out
        cls.cache[key] = list_out
    
    def _stream_completion(self):
        pass

    # Define a function to add a call to the deque
    def add_call(self):
        # Get the current timestamp in seconds
        now = time.time()
        # Append the timestamp to the right of the deque
        self.call_history.append(now)

    # Define a function to count the calls in the last 60 seconds
    def count_calls(self):
        # Get the current timestamp in seconds
        now = time.time()
        # Remove the timestamps that are older than 60 seconds from the left of the deque
        while self.call_history and self.call_history[0] < now - 60:
            self.call_history.popleft()
        # Return the length of the deque as the number of calls
        return len(self.call_history)

    def _library_call(self, **kwargs):
        """ Call the OpenAI API using the python package.

        Note that is uses the local auth token, and does not rely on the openai one.
        """
        prev_key = openai.api_key
        assert self.token is not None, "You must provide an OpenAI API key to use the OpenAI LLM. Either pass it in the constructor, set the OPENAI_API_KEY environment variable, or create the file ~/.openai_api_key with your key in it."
        openai.api_key = self.token
        if self.chat_mode:
            kwargs['messages'] = prompt_to_messages(kwargs['prompt'])
            del kwargs['prompt']
            del kwargs['echo']
            del kwargs['logprobs']
            # print(kwargs)
            out = openai.ChatCompletion.create(**kwargs)
            out = add_text_to_chat_mode(out)
        else:
            out = openai.Completion.create(**kwargs)
        openai.api_key = prev_key
        return out

    def _rest_call(self, **kwargs):
        """ Call the OpenAI API using the REST API.
        """

        # Define the request headers
        headers = copy.copy(self._rest_headers)
        if self.token is not None:
            headers['Authorization'] = f"Bearer {self.token}"

        # Define the request data
        stream = kwargs.get("stream", False)
        data = {
            "prompt": kwargs["prompt"],
            "max_tokens": kwargs.get("max_tokens", None),
            "temperature": kwargs.get("temperature", 0.0),
            "top_p": kwargs.get("top_p", 1.0),
            "n": kwargs.get("n", 1),
            "stream": stream,
            "logprobs": kwargs.get("logprobs", None),
            'stop': kwargs.get("stop", None),
            "echo": kwargs.get("echo", False)
        }
        if self.chat_mode:
            data['messages'] = prompt_to_messages(data['prompt'])
            del data['prompt']
            del data['echo']
            del data['stream']

        # Send a POST request and get the response
        response = requests.post(self.endpoint, headers=headers, json=data, stream=stream)
        if response.status_code != 200:
            raise Exception("Response is not 200: " + response.text)
        if stream:
            return self._rest_stream_handler(response)
        else:
            response = response.json()
        if self.chat_mode:
            response = add_text_to_chat_mode(response)
        return response
        
    def _rest_stream_handler(self, response):
        for line in response.iter_lines():
            text = line.decode('utf-8')
            if text.startswith('data: '):
                text = text[6:]
                if text == '[DONE]':
                    break
                else:
                    yield json.loads(text)
    
    def encode(self, string, fragment=True):
        # note that is_fragment is not used used for this tokenizer
        return self._tokenizer.encode(string)
    
    def decode(self, tokens, fragment=True):
        return self._tokenizer.decode(tokens)


# Define a deque to store the timestamps of the calls
class OpenAISession(LLMSession):
    async def __call__(self, prompt, stop=None, stop_regex=None, temperature=None, n=1, max_tokens=1000, logprobs=None, top_p=1.0, echo=False, logit_bias=None, token_healing=None, pattern=None, stream=False, cache_seed=0, caching=None):
        """ Generate a completion of the given prompt.
        """

        assert token_healing is None or token_healing is False, "The OpenAI API does not support token healing! Please either switch to an endpoint that does, or don't use the `token_healing` argument to `gen`."

        # set defaults
        if temperature is None:
            temperature = self.llm.temperature

        # get the arguments as dictionary for cache key generation
        args = locals().copy()

        assert not pattern, "The OpenAI API does not support Guidance pattern controls! Please either switch to an endpoint that does, or don't use the `pattern` argument to `gen`."
        assert not stop_regex, "The OpenAI API does not support Guidance stop_regex controls! Please either switch to an endpoint that does, or don't use the `stop_regex` argument to `gen`."

        # define the key for the cache
        key = self._cache_key(args)
        
        # allow streaming to use non-streaming cache (the reverse is not true)
        if key not in self.llm.__class__.cache and stream:
            args["stream"] = False
            key1 = self._cache_key(args)
            if key1 in self.llm.__class__.cache:
                key = key1
        
        # check the cache
        if key not in self.llm.__class__.cache or (caching is not True and not self.llm.caching) or caching is False:

            # ensure we don't exceed the rate limit
            while self.llm.count_calls() > self.llm.max_calls_per_min:
                await asyncio.sleep(1)

            fail_count = 0
            while True:
                try_again = False
                try:
                    self.llm.add_call()
                    call_args = {
                        "model": self.llm.model_name,
                        "prompt": prompt,
                        "max_tokens": max_tokens,
                        "temperature": temperature,
                        "top_p": top_p,
                        "n": n,
                        "stop": stop,
                        "logprobs": logprobs,
                        "echo": echo,
                        "stream": stream
                    }
                    if logit_bias is not None:
                        call_args["logit_bias"] = logit_bias
                    out = self.llm.caller(**call_args)

                except openai.error.RateLimitError:
                    await asyncio.sleep(3)
                    try_again = True
                    fail_count += 1
                
                if not try_again:
                    break

                if fail_count > self.llm.max_retries:
                    raise Exception(f"Too many (more than {self.llm.max_retries}) OpenAI API RateLimitError's in a row!")

            if stream:
                return self.llm.stream_then_save(out, key)
            else:
                self.llm.__class__.cache[key] = out
        
        # wrap as a list if needed
        if stream:
            if isinstance(self.llm.__class__.cache[key], list):
                return self.llm.__class__.cache[key]
            return [self.llm.__class__.cache[key]]
        
        return self.llm.__class__.cache[key]
    
# class OpenAISession(AsyncOpenAISession):
#     def __call__(self, *args, **kwargs):
#         return self._loop.run_until_complete(super().__call__(*args, **kwargs))


from ._llm import LLM

class Mock(LLM):
    """ Mock class for testing.
    """
    def __init__(self, output=None):
        """ Initialize the mock class.
        
        Parameters
        ----------
        output : object
            The output of the mock class. If a list is provided, the mock
            class will return the next item in the list each time it is
            called. Otherwise, the same output will be returned each time.
        """
        if output is None:
            output = [f"mock output {i}" for i in range(100)]
        self.output = output
        self.index = 0
        self._tokenizer = MockTokenizer()

    def __call__(self, *args, n=1, stream=False, **kwargs):
        choices = []
        for i in range(n):
            if isinstance(self.output, list):
                out = self.output[self.index]
                self.index += 1
            else:
                out = self.output

            choices.append({"text": out})

        out = {"choices": choices}

        if stream:
            return [out]
        else:
            return out
        
    def role_start(self, role):
        return "<|im_start|>"+role+"\n"
    
    def role_end(self, role=None):
        return "<|im_end|>"
    
class MockTokenizer():
    def __init__(self):
        pass

    def encode(self, text):
        return [s for s in text.encode("utf-8")]
    
    def decode(self, ids):
        return "".join([chr(i) for i in ids])


from ._role import role

async def user(hidden=False, _parser_context=None):
    ''' A chat role block for the 'user' role.

    This is just a shorthand for {{#role 'user'}}...{{/role}}.

    Parameters
    ----------
    hidden : bool
        Whether to include the assistant block in future LLM context. 
    '''
    return await role(name="user", hidden=hidden, _parser_context=_parser_context)
user.is_block = True


import os
import time
import collections
import regex
import pygtrie
import queue
import threading
import logging
from ._llm import LLM, LLMSession, SyncSession

class Transformers(LLM):
    """ A HuggingFace transformers language model with Guidance support.
    """

    cache = LLM._open_cache("_transformers.diskcache")

    def __init__(self, model=None, tokenizer=None, caching=True, token_healing=True, acceleration=True, temperature=0.0, device=None, device_map=None, role_start=None, role_end=None):
        super().__init__()

        # fill in default model value
        if model is None:
            model = os.environ.get("TRANSFORMERS_MODEL", None)
        if model is None:
            try:
                with open(os.path.expanduser('~/.transformers_model'), 'r') as file:
                    model = file.read().replace('\n', '')
            except:
                pass

        self.model_obj, self._tokenizer = self._model_and_tokenizer(model, tokenizer, device_map)
        self._generate_call = self.model_obj.generate

        self.model_name = model
        self.caching = caching
        self.current_time = time.time()
        self.call_history = collections.deque()
        self.temperature = temperature
        self.token_healing = token_healing
        self.acceleration = acceleration
        if device is not None: # set the device if requested
            self.model_obj = self.model_obj.to(device)
        self.device = self.model_obj.device # otherwise note the current device

        self._prefix_ids = [self._tokenizer.bos_token_id, 100] # token ids that we use to decode tokens after a prefix
        self._prefix_str = self._tokenizer.decode(self._prefix_ids, fragment=False)

        self._token_prefix_map = self._build_token_prefix_map(model)

    def prefix_matches(self, prefix):
        """ Return the list of tokens that match the given prefix.
        """
        return [v for arr in self._token_prefix_map.values(prefix=prefix) for v in arr]

    def encode(self, string, fragment=True, **kwargs):

        if fragment:
            string = self._prefix_str + string

        if "return_tensors" in kwargs:
            out = self._tokenizer(string, **kwargs)
        else:
            out = self._tokenizer.encode(string, **kwargs)
        
        # remove the start token when we are encoding a suffix
        if fragment:
            if out[1] == self._tokenizer.bos_token_id: # sometime the tokenizer adds an extra start token
                out = out[3:]
            else:
                out = out[2:]
        
        return out
    
    # def role_start(self, role):
    #     """ The starting role tag for chat models.

    #     #TODO Right now this just assumes the StableLM syntax, but this should be expanded later.
    #     """
    #     return "<|"+role.upper()+"|>"
    
    # def role_end(self, role=None):
    #     return ""
    
    def decode(self, tokens, fragment=True, **kwargs):

        # if the last token is the end of string token, or the first is a start of string we remove it because it cause odd spacing decoding of fragments
        add_eos = ""
        add_bos = ""
        if fragment:
            if len(tokens) > 0 and tokens[-1] == self._tokenizer.eos_token_id:
                add_eos = self._tokenizer.eos_token
                tokens = tokens[:-1]
            if len(tokens) > 0 and tokens[0] == self._tokenizer.bos_token_id:
                add_bos = self._tokenizer.bos_token
                tokens = tokens[1:]
        
        # Decode the string corresponding to a single suffix token.
        # Note that we need to decode after the start token for sentence-piece tokenizers so that white space is preserved
        if fragment:
            return add_bos + self._tokenizer.decode(self._prefix_ids + list(tokens))[len(self._prefix_str):] + add_eos
        else:
            return add_bos + self._tokenizer.decode(tokens, **kwargs) + add_eos

    def _build_token_prefix_map(self, model_name):
        """ Build a map from token to index.
        """
        token_map = pygtrie.CharTrie()
        for i in range(self._tokenizer.vocab_size):
            s = self.decode([i])
            if s in token_map:
                token_map[s].append(i) # handle duplicate token encodings... (GPT2 BPE has this oddly enough)
            else:
                token_map[s] = [i]

        return token_map

    def _model_and_tokenizer(self, model, tokenizer, device_map):

        # make sure transformers is installed
        try:
            import transformers
        except:
            raise Exception("Please install transformers with `pip install transformers` in order to use guidance.llms.Transformers!")

        # intantiate the model and tokenizer if needed
        if isinstance(model, str):
            if tokenizer is None:
                tokenizer = transformers.AutoTokenizer.from_pretrained(model, device_map=device_map)
            model = transformers.AutoModelForCausalLM.from_pretrained(model, device_map=device_map)
        
        assert tokenizer is not None, "You must give a tokenizer object when you provide a model object (as opposed to just a model name)!"
            
        return model, tokenizer

    def session(self, asynchronous=False):
        if asynchronous:
            return TransformersSession(self)
        else:
            return SyncSession(TransformersSession(self))


class TransformersSession(LLMSession):
    def __init__(self, llm):
        super().__init__(llm)
        
        self._past_key_values = None
        self._prefix_cache = []
    
    def __enter__(self):

        # we only need decorators if we are using token acceleration
        if self.llm.acceleration:

            # decorate the prep step to preserve the initial past key values we have passed
            def prep_step_decorator(method):
                def decorate_prep_step(input_ids, **kwargs):

                    # if we are extending the input ids with the cached tokens then
                    # don't pass past key values to the input prep step, otherwise it
                    # would delete all but the last input_ids, and we have already removed
                    # the correct prefix from the input_ids (which is not always all but the last one)
                    if len(self._prefix_cache) > 0:
                        
                        kwargs["past"] = None
                        input_ids = input_ids[:,len(self._prefix_cache):]
                        # if "attention_mask" in kwargs:
                        #     kwargs["attention_mask"] = kwargs["attention_mask"][:,len(self._prefix_cache):]
                        model_kwargs = method(input_ids, **kwargs)

                        # provide the past key values for the actual model call
                        model_kwargs["past_key_values"] = self._past_key_values
                        model_kwargs["position_ids"] = model_kwargs["position_ids"][:,len(self._prefix_cache):] # and update position ids

                        # we only need to do this first time, after that the past key values will
                        # be up until the last token, just like transformer models normally expect
                        # so we can clear our cache and let transformers cache like normal
                        self._prefix_cache = [] # this will get refilled once the generate call is done
                    
                        return model_kwargs
                    else:
                        return method(input_ids, **kwargs)
                decorate_prep_step.__func__ = method.__func__ # make us still look like a bound method
                return decorate_prep_step
            if getattr(self.llm.model_obj, "_orig_prepare_method", None) is None:
                self.llm.model_obj._orig_prepare_method = self.llm.model_obj.prepare_inputs_for_generation
            self.llm.model_obj.prepare_inputs_for_generation = prep_step_decorator(self.llm.model_obj._orig_prepare_method)

            # decorate the update step to save the past key values
            def update_step_decorator(method):
                def decorate_update_step(outputs, *args, **kwargs):

                    # save the past key values
                    self._past_key_values = outputs.past_key_values

                    return method(outputs, *args, **kwargs)
                return decorate_update_step
            if getattr(self.llm.model_obj, "_orig_update_method", None) is None:
                self.llm.model_obj._orig_update_method = self.llm.model_obj._update_model_kwargs_for_generation
            self.llm.model_obj._update_model_kwargs_for_generation = update_step_decorator(self.llm.model_obj._orig_update_method)

        return self

    # def __call__(self, *args, **kwargs):
    #     return self.__call__(*args, **kwargs)
    
    async def __call__(self, prompt, stop=None, stop_regex=None, temperature=None, n=1, max_tokens=1000, logprobs=None, top_p=1.0, echo=False, logit_bias=None, token_healing=None, pattern=None, stream=False, cache_seed=0, caching=None):
        """ Generate a completion of the given prompt.
        """
        
        # fill in defaults
        if temperature is None:
            temperature = self.llm.temperature
        if token_healing is None:
            token_healing = self.llm.token_healing

        # generate the cache key
        key = self._cache_key(locals())

        # set the stop patterns
        if stop is not None:
            if isinstance(stop, str):
                stop_regex = [regex.escape(stop)]
            else:
                stop_regex = [regex.escape(s) for s in stop]
        if isinstance(stop_regex, str):
            stop_regex = [stop_regex]
        if stop_regex is None:
            stop_regex = []
        stop_regex.append(self.llm._tokenizer.eos_token) # make sure the end of sequence token is always included

        # handle caching
        if key not in self.llm.cache or (caching is not True and not self.llm.caching) or caching is False:
            import transformers
            # import torch
            # encode the prompt
            encoded = self.llm.encode([prompt for _ in range(n)], return_tensors="pt", fragment=False)
            if self.llm.device is not None:
                encoded = encoded.to(self.llm.device)
            input_ids = encoded["input_ids"]
            attention_mask = encoded["attention_mask"]
            model_config = self.llm.model_obj.config

            # ensure that we are extending a common sequence batch (our token healing assumes this right now)
            assert (input_ids[0,-1] == input_ids[:,-1]).all(), "The current token healing implementation assumes that batches are reps of the same sequence!"

            last_token_str = ""
            processors = []
            stoppers = []

            # save what the prompt looks like when coded and then decoded (this captures added start tokens, etc.)
            coded_prompt = self.llm.decode(input_ids[0])

            # setup token healing
            if token_healing:
                # pop off the last token since we will regen it
                last_token_id = input_ids[0][-1]
                last_token_str = self.llm.decode([last_token_id])
                healer = TokenHealingLogitsProcessor(self.llm, model_config.vocab_size, last_token_str)
                if healer.should_bias:
                    input_ids = input_ids[:,:-1]
                    attention_mask = attention_mask[:,:-1]
                    max_tokens += 1 # add one for the token we regen for token healing
                    processors.append(healer)
                else:
                    last_token_str = ""

            # make sure we don't run off the end of the model
            max_context = (getattr(model_config, "max_sequence_length", None) or getattr(model_config, "max_seq_len", None) or getattr(model_config, "n_positions", None) or getattr(model_config, "max_position_embeddings"))
            if max_tokens + len(input_ids[0]) > max_context:
                max_tokens = max_context - len(input_ids[0])

            # find how much of the prompt is cached
            for prefix_match_len, token in enumerate(input_ids[0]):
                if prefix_match_len >= len(self._prefix_cache) or token != self._prefix_cache[prefix_match_len]:
                    break

            # trim the cache to what we can use
            if prefix_match_len > 0 and prefix_match_len < len(self._prefix_cache):
                self._past_key_values = tuple((key[:,:,:prefix_match_len,:],value[:,:,:prefix_match_len,:]) for key,value in self._past_key_values) # TODO: this is specific to the GPT2 tensor layout
                self._prefix_cache = self._prefix_cache[:prefix_match_len]

            # see if we need to returns the scores
            # output_scores = logprobs is not None and logprobs > 0

            # position_ids = torch.arange(prefix_match_len, input_ids.shape[-1], dtype=torch.long).unsqueeze(0)
                
            # trim input ids that we will pull from the cache instead of computing keys and values for
            # input_ids = input_ids[:,prefix_match_len:]

            # add support for pattern guidance
            if pattern is not None:
                processors.append(RegexLogitsProcessor(pattern, stop_regex, self.llm.decode, model_config.vocab_size, temperature == 0, len(coded_prompt), self.llm._tokenizer.eos_token_id))

            if stop_regex is not None:
                stoppers.append(RegexStoppingCriteria(stop_regex, self.llm.decode, len(coded_prompt)))

            # a streamer to handle potentially partial output
            streamer = TransformersStreamer(
                input_ids=input_ids,
                stop_regex=stop_regex,
                last_token_str=last_token_str,
                coded_prompt=coded_prompt,
                llm=self.llm,
                max_new_tokens=max_tokens,
                lobprobs=logprobs
            )

            # the args for the transformers generate call
            generate_args = dict(
                inputs=input_ids,
                attention_mask=attention_mask,
                # position_ids=position_ids,
                temperature=temperature,
                max_new_tokens=max_tokens,
                top_p=top_p,
                pad_token_id=model_config.pad_token_id if model_config.pad_token_id is not None else self.llm._tokenizer.eos_token_id,
                logits_processor=transformers.LogitsProcessorList(processors),
                stopping_criteria=transformers.StoppingCriteriaList(stoppers),
                # past_key_values=self._past_key_values,
                output_scores=logprobs is not None and logprobs > 0,
                return_dict_in_generate=True
            )

            # override the model config for do_sample when the temperature requires it
            do_sample = getattr(self.llm.model_obj.config, "do_sample", None)
            if do_sample is True and temperature == 0:
                generate_args["do_sample"] = False
            elif do_sample is False and temperature > 0:
                generate_args["do_sample"] = True

            # if we are streaming then we need to run the inference process in a separate thread
            if stream:
                generate_args["streamer"] = streamer
                thread = threading.Thread(target=self.llm._generate_call, kwargs=generate_args)
                thread.start()
                return self._stream_then_save(streamer, key, thread)

            # if we are not streaming we still manually use the streamer for consistency
            else:
                generated_sequence = self.llm._generate_call(**generate_args)
                streamer.put(generated_sequence)
                self.llm.cache[key] = streamer.__next__()
                self._update_prefix_cache(streamer)
        return self.llm.cache[key]
    
    def _update_prefix_cache(self, streamer):
        # note what we now have cached and ready for our next call in this session
        if self._past_key_values and len(streamer.generated_sequence) == 1:
            self._prefix_cache = streamer.generated_sequence[0][:self._past_key_values[0][0].shape[2]] # self._past_key_values is already saved, this just aligns with it

    def _stream_then_save(self, streamer, key, thread):
        list_out = []
        for out in streamer:
            list_out.append(out)
            yield out
        thread.join() # clean up the thread
        self.llm.cache[key] = list_out
        self._update_prefix_cache(streamer)

    def __exit__(self, exc_type, exc_value, traceback):
        """ Restore the model to its original state by removing monkey patches.
        """
        if getattr(self.llm.model_obj, "_orig_prepare_method", None) is not None:
            self.llm.model_obj.prepare_inputs_for_generation = self.llm.model_obj._orig_prepare_method
            del self.llm.model_obj._orig_prepare_method
        if getattr(self.llm.model_obj, "_orig_update_method", None) is not None:
            self.llm.model_obj._update_model_kwargs_for_generation = self.llm.model_obj._orig_update_method
            del self.llm.model_obj._orig_update_method
        return False



class TokenHealingLogitsProcessor():
    """ Token healing.

    When we tokenize the prompt the last token we get is not the last token we would
    have gotten if the prompt + generation was concatented and then tokenized. This
    is not good because it does not align with the pretraining of the model, so
    we "heal" this boundary by backing up one token and then forcing the first token
    generated to start with the prefix of the last token in the prompt. This could
    result in the same token as the end of the prompt, or another longer one.
    """

    def __init__(self, model, vocab_size, last_token_str, bias_value=50.):
        """ Build a new TokenHealingLogitsProcessor.

        Note that bias value is in score space (log-odds normally) and should be
        enough to ensure those tokens are the only ones used. But not so high
        as to destroy numerical precision.
        """
        import torch

        try:
            allowed_first_tokens = model.prefix_matches(last_token_str)
            assert len(allowed_first_tokens) > 0, "Error in token healing map! No match found for: `"+last_token_str+"`"
        except KeyError:
            # this must be a special token outside the vocab, so we assume it does not have any valid extensions
            allowed_first_tokens = []
        
        # if we have multiple possible completions past the last token, then biasing is needed
        if len(allowed_first_tokens) > 1:
            self.first_token_mask = torch.zeros(vocab_size)
            self.first_token_mask.scatter_(0, torch.tensor(allowed_first_tokens), bias_value)
            if model.device is not None:
                self.first_token_mask = self.first_token_mask.to(model.device)
            self.should_bias = True
        
        # otherwise we have nothing to do (the last token is already unique)
        else:
            self.should_bias = False

    def __call__(self, input_ids, scores):

        # we only bias the first token generated
        if not self.should_bias:
            return scores
        self.should_bias = False
        
        # make only allowed tokens possible
        return scores + self.first_token_mask
    
class RegexLogitsProcessor():
    """ Pattern guiding.
    
    Guide generation to match a regular expression.
    TODO: currently slow, could be made much faster by doing rejection sampling inline with the sampling/greedy process.
    """

    def __init__(self, pattern, stop_regex, decode, vocab_size, is_greedy, prefix_length, eos_token_id, max_consider=10000, bias_value=30.):
        """ Build a new TokenHealingLogitsProcessor.

        Parameters
        ----------
        pattern : str
            The regex pattern we are seeking to match.
        stop_regex : str
            The stop regex that is allowed to come after this pattern.
        decode : function
            The token decoding function.
        vocab_size : int
            The size of the vocabulary.
        is_greedy : bool
            The token selection mode currently in use. We need to know this so we can
            effectively take over that sampling process inside this logit processor.
        max_consider : int
            How many top values to bias. Note that we could remove this option once this
            processor is performance optimized (by integrating it into the sampling/greedy process).
        bias_value : float
            The bias value is in score space (log-odds normally) and should be
            enough to ensure those tokens are the only ones used. But not so high
            as to destroy numerical precision.
        eos_token_id : int
            The end of the stop token of the model.
        """
        import torch
        
        if stop_regex is not None:
            pattern += "(" + "|".join(stop_regex) + ")?"
        stop_token_str = regex.escape(decode([eos_token_id]))
        pattern += "(" + stop_token_str + ")?"
        self.pattern = regex.compile(pattern)
        self.decode = decode
        self.is_greedy = is_greedy
        self.prefix_length = prefix_length
        self.max_consider = max_consider
        self.bias_value = bias_value
        self.bias_vector = torch.zeros(vocab_size)
        self.current_strings = None
        self.current_length = 0
        self.forced_chars = 0

    def __call__(self, input_ids, scores):
        import torch

        # extend our current strings
        if self.current_strings is None:
            self.current_strings = ["" for _ in range(len(input_ids))]
        for i in range(len(self.current_strings)):
            self.current_strings[i] += self.decode(input_ids[i][self.current_length:])

        # trim off the prefix string so we don't look for stop matches in the prompt
        if self.current_length == 0:
            self.forced_chars = self.prefix_length - len(self.current_strings[i]) # account for token healing forced prefixes
            for i in range(len(self.current_strings)):
                self.current_strings[i] = self.current_strings[i][self.prefix_length:]

        self.current_length = len(input_ids[0])
        
        # compute the bias values
        self.bias_vector[:] = 0
        sort_inds = torch.argsort(scores, 1, True)
        for i in range(min(sort_inds.shape[1], self.max_consider)):
            m = self.pattern.fullmatch((self.current_strings[0] + self.decode([sort_inds[0,i]]))[self.forced_chars:], partial=True) # partial means we don't match currently but might as the string grows
            if m:
                self.bias_vector[sort_inds[0,i]] = self.bias_value
                if self.is_greedy:
                    break # we are done if we are doing greedy sampling and we found the top valid hit
        
        # make only allowed tokens
        return scores + self.bias_vector.to(scores.device)

class RegexStoppingCriteria():
    def __init__(self, stop_pattern, decode, prefix_length):
        if isinstance(stop_pattern, str):
            self.stop_patterns = [regex.compile(stop_pattern)]
        else:
            self.stop_patterns = [regex.compile(pattern) for pattern in stop_pattern]
        self.prefix_length = prefix_length
        self.decode = decode
        self.current_strings = None
        self.current_length = 0

    def __call__(self, input_ids, scores, **kwargs):

        # extend our current strings
        if self.current_strings is None:
            self.current_strings = ["" for _ in range(len(input_ids))]
        for i in range(len(self.current_strings)):
            self.current_strings[i] += self.decode(input_ids[i][self.current_length:])
        
        # trim off the prefix string so we don't look for stop matches in the prompt
        if self.current_length == 0:
            for i in range(len(self.current_strings)):
                self.current_strings[i] = self.current_strings[i][self.prefix_length:]
        
        self.current_length = len(input_ids[0])
        
        # check if all of the strings match a stop string (and hence we can stop the batch inference)
        all_done = True
        for i in range(len(self.current_strings)):
            found = False
            for s in self.stop_patterns:
                if s.search(self.current_strings[i]):
                    found = True
            if not found:
                all_done = False
                break
        
        return all_done

class TransformersStreamer():
    def __init__(self, input_ids, stop_regex, last_token_str, coded_prompt, llm, max_new_tokens, lobprobs, timeout=None):
        self.timeout = timeout
        self.input_ids = input_ids
        self.stop_regex = stop_regex
        self.logprobs = lobprobs
        self.last_token_str = last_token_str
        # self.coded_prompt = coded_prompt
        self.llm = llm
        self.max_total_tokens = max_new_tokens + len(input_ids[0])
        coded_prompt = coded_prompt[:len(coded_prompt)-len(last_token_str)] # strip off the last token which will be regenerated
        self.str_pos = [len(coded_prompt) + len(self.last_token_str) for i in range(len(self.input_ids))]
        self.out_queue = queue.Queue()
        self.sequence_pos = [len(self.input_ids[0]) for i in range(len(self.input_ids))]
        self.generated_sequence = [[] for i in range(len(self.input_ids))]
        self.generated_scores = [[] for i in range(len(self.input_ids))]
        self.generated_string = [coded_prompt for i in range(len(self.input_ids))]
        self.prefix_cache = []

    def put(self, token_obj):

        import torch
        if isinstance(token_obj, torch.Tensor):
            new_tokens = token_obj
        else:
            new_tokens = token_obj['sequences']
        
        # if we are given a single sequence, then make it a batch of size 1
        if len(new_tokens.shape) == 1:
            new_tokens = new_tokens.unsqueeze(0)
        
        # extract the scores if we are given them (and format them to be the same shape as the tokens)
        if self.logprobs:
            assert len(new_tokens) == 1, "logprobs are not supported for batched generation right now in guidance.llms.Transformers"
            new_scores = list(token_obj['scores'])
            len_diff = len(new_tokens[0]) - len(new_scores)
            if len_diff > 0:
                new_scores = [None for i in range(len_diff)] + new_scores
            new_scores = [new_scores]
        
        out = {"choices": [None for i in range(len(self.input_ids))]}
        put_data = False
        for i in range(len(self.input_ids)):
            self.generated_sequence[i].extend(list(new_tokens[i]))
            if self.logprobs:
                self.generated_scores[i].extend(list(new_scores[i]))

            if self.sequence_pos[i] < len(self.generated_sequence[i]):
                display_tokens = list(self.generated_sequence[i][self.sequence_pos[i]:])
                val = self.llm.decode(display_tokens)#[self.llm._prefix_token_id] + display_tokens)[len(self.llm._prefix_token):]
                self.generated_string[i] += val
                
                if self.str_pos[i] < len(self.generated_string[i]):
                    
                    display_logprobs = None
                    if self.logprobs:
                        display_scores = self.generated_scores[i][self.sequence_pos[i]:]
                        display_logprobs = []
                        for k in range(len(display_scores)):
                            top_inds = display_scores[k][0].argsort(descending=True)[:self.logprobs] # TODO: verify the [0] is always correct
                            display_logprobs.append({self.llm.decode([j]): display_scores[k][0][j] for j in top_inds})

                    val = self.generated_string[i][self.str_pos[i]:]
                    finish_reason = None
                    
                    # check why we stopped
                    stop_pos = len(val) + 1
                    if len(self.generated_sequence[i]) >= self.max_total_tokens:
                        finish_reason = "length"
                    elif self.generated_sequence[i][-1] == self.llm._tokenizer.eos_token_id:
                        finish_reason = "endoftext"
                        stop_pos = len(val) - len(self.llm.decode([self.llm._tokenizer.eos_token_id]))

                    # trim off the stop regex matches if needed
                    found_partial = False
                    if self.stop_regex is not None and (finish_reason is None or len(self.input_ids) > 1):
                        stop_regex_obj = [regex.compile(s) for s in self.stop_regex]
                        for s in stop_regex_obj:
                            m = s.search(val, partial=True)
                            if m:
                                span = m.span()
                                if span[1] > span[0]:
                                    if m.partial: # we might be starting a stop sequence, so we can't emit anything yet
                                        found_partial = True
                                        break
                                    else:
                                        stop_pos = min(span[0], stop_pos)

                    # record the reason we stopped (if we have stopped)
                    if stop_pos <= len(val):
                        finish_reason = "stop"
                    
                    if not found_partial:
                        out["choices"][i] = {
                            "text": val[:stop_pos],
                            "finish_reason": finish_reason,
                            "logprobs": {"token_healing_prefix": self.last_token_str, "top_logprobs": display_logprobs}
                        }
                        self.str_pos[i] = len(self.generated_string[i])
                        put_data = True
                self.sequence_pos[i] = len(self.generated_sequence[i])
        if put_data:
            self.out_queue.put(out)

    def end(self):

        # make sure we have flushed all of the data
        for i in range(len(self.input_ids)):
            assert self.str_pos[i] >= len(self.generated_string[i]), "Not all data was flushed, this means generation stopped for an unknown reason!"
        
        self.out_queue.put(None)

    def __iter__(self):
        return self

    def __next__(self):
        value = self.out_queue.get(timeout=self.timeout)
        if value is None:
            raise StopIteration()
        else:
            return value



import itertools
import pygtrie
import numpy as np

async def select(variable_name="selected", options=None, logprobs=None, _parser_context=None):
    ''' Select a value from a list of choices.

    Parameters
    ----------
    variable_name : str
        The name of the variable to set with the selected value.
    options : list of str or None
        An optional list of options to select from. This argument is only used when select is used in non-block mode.
    logprobs : str or None
        An optional variable name to set with the logprobs for each option.
    '''
    parser = _parser_context['parser']
    block_content = _parser_context['block_content']
    parser_prefix = _parser_context['parser_prefix']
    partial_output = _parser_context['partial_output']

    if block_content is None:
        assert options is not None, "You must provide an options list like: {{select 'variable_name' options}} when using the select command in non-block mode."
    else:
        assert len(block_content) > 1, "You must provide at least one two options to the select block command."
        assert options is None, "You cannot provide an options list when using the select command in block mode."

    if options is None:
        options = [block_content[0].text]
        for i in range(1, len(block_content), 2):
            assert block_content[i].text == "{{or}}"
            options.append(block_content[i+1].text)

    option_tokens = [parser.program.llm.encode(option) for option in options]
    ids_used = set(itertools.chain.from_iterable(option_tokens))

    # build a trie of the options
    token_map = pygtrie.CharTrie()
    for i,option in enumerate(options):
        token_map[option] = i
    
    async def rec_select(current_prefix):
        """ This returns a dictionary of scores for each option (keyed by the option index).
        """

        # this is the dictionary of logprobs for each option we will return
        # note that the logprobs are just for this branch point and below in the decision tree
        # logprobs_out = {option: -100 for option in options}

        # find which select options are possible
        try:
            extension_options = token_map.items(prefix=current_prefix)
        except KeyError:
            return {}

        # this is the dictionary of logprobs for each option we will return
        # note that the logprobs are just for this branch point and below in the decision tree
        logprobs_out = {option[0]: -1000 for option in extension_options}

        # extend the prefix with the longest common prefix among the valid options
        # we also stop early if we have one option
        if len(extension_options) == 1:
            logprobs_out[extension_options[0][0]] = 0 # probability of 1.0 that we will select the only valid option
            return logprobs_out
        else:
            for i in range(len(current_prefix), min([len(o[0]) for o in extension_options])):
                if len(set([o[0][i] for o in extension_options])) > 1:
                    break
            current_prefix = extension_options[0][0][:i]

        # extend the prefix by one token using the model
        gen_obj = await parser.llm_session(
            parser_prefix + current_prefix,
            max_tokens=1,
            # logit_bias={str(id): 50 for id in ids_used},
            logprobs=10,
            cache_seed=0
        )
        logprobs_result = gen_obj["choices"][0]["logprobs"]
        top_logprobs = logprobs_result["top_logprobs"][0]

        remove_prefix = len(logprobs_result.get("token_healing_prefix", ""))

        # for each possible next token, see if it grows the prefix in a valid way
        for token_str,logprob in top_logprobs.items():
            if len(token_str[remove_prefix:]) > 0:
                sub_logprobs = await rec_select(current_prefix + token_str[remove_prefix:])
                for k in sub_logprobs:
                    logprobs_out[k] = sub_logprobs[k] + logprob

        # if we did token healing and did not extend past our prefix we need to consider the next token
        # TODO: when returning all logprobs we need to consider all the options, which means we should
        # force the model to not token heal and see what would have happened then on the next token...
        first_token_str = max(top_logprobs, key=top_logprobs.get)
        if len(logprobs_result["top_logprobs"]) > 1 and len(first_token_str) == remove_prefix:            
            top_logprobs = logprobs_result["top_logprobs"][1]
            for token_str,logprob in top_logprobs.items():
                sub_logprobs = await rec_select(current_prefix + token_str)
                for k in sub_logprobs:

                    # compute the probability of a logical OR between the new extension and the previous possible ones
                    p1 = np.exp(logprobs_out[k])
                    p2 = np.exp(sub_logprobs[k] + logprob)
                    or_prob = p1 + p2 - p1*p2
                    logprobs_out[k] = np.log(or_prob)

        return logprobs_out
        
    # recursively compute the logprobs for each option
    option_logprobs = await rec_select("") 

    selected_option = max(option_logprobs, key=option_logprobs.get)
    parser.set_variable(variable_name, selected_option)
    if logprobs is not None:
        parser.set_variable(logprobs, option_logprobs)
    
    if max(option_logprobs.values()) <= -1000:
        raise ValueError("No valid option generated in #select, this could be fixed if we used a tokenizer and forced the LM to use a valid option! The top logprobs were" + str(top_logprobs))
    
    partial_output(selected_option)

select.is_block = True


def set(name, value=None, hidden=None, _parser_context=None):
    ''' Set the value of a variable or set of variables.

    Parameters
    ----------
    name : str or dict
        If a string, the name of the variable to set. If a dict, the keys are the variable names and the values are the values to set.
    value : str, optional
        The value to set the variable to. Only used if `name` is a string.
    hidden : bool, optional
        If True, the variable will be set but not printed in the output.
    '''
    parser = _parser_context['parser']

    if not parser.executing:
        return ""

    if isinstance(name, dict):
        assert hidden is not False, "hidden cannot be False if setting multiple variables!"
        for k, v in name.items():
            parser.set_variable(k, v)
        out = ""
        for k, v in name.items():
            if isinstance(v, str):
                if "\n" in v:
                    v = f'"""{v}"""'
                elif '"' in v:
                    v = f"'{v}'"
                else:
                    v = f'"{v}"'
            out += f" {k}={v}"
        out += ""
        return "{{!--GMARKER_set$" + out + "$--}}"
    else:
        parser.set_variable(name, value)
        if hidden is not True:
            return value
        else:
            out = "{{set "+name+"=" + str(value) + "}}"
            return "{{!--GMARKER_set$" + out + "$--}}"


import inspect
import re
import uuid
from .._utils import strip_markers

async def geneach(list_name, stop=None, max_iterations=100, min_iterations=0, num_iterations=None, hidden=False, join="", single_call=False, single_call_temperature=0.0, single_call_max_tokens=500, single_call_top_p=1.0, _parser_context=None):
    ''' Generate a potentially variable length list of items using the LLM.

    Parameters
    ----------
    list_name : str
        The name of the variable to save the generated list to.
    stop : str or list of str
        A string or list of strings that will stop the generation of the list. For example if stop="</ul>"
        then the list will be generated until the first "</ul>" is generated.
    max_iterations : int
        The maximum number of items to generate.
    min_iterations : int
        The minimum number of items to generate.
    num_iterations : int
        The exact number of items to generate (this overrides max_iterations and min_iterations).
    hidden : bool
        If True, the generated list items will not be added to the LLMs input context. This means that each
        item will be generated independently of the others. Note that if you use hidden=True you must also
        set num_iterations to a fixed number (since without adding items the context there is not way for the
        LLM to know when to stop on its own).
    join : str
        A string to join the generated items with.
    single_call : bool
        This is an option designed to make look generation more convienent for LLMs that don't support guidance
        acceleration. If True, the LLM will be called once to generate the entire list. This only works if the
        LLM has already been prompted to generate content that matches the format of the list. After the single
        call, the generated list variables will be parsed out of the generated text using a regex. (note that only
        basic template tags are supported in the list items when using single_call=True).
    single_call_temperature : float
        Only used with single_call=True. The temperature to use when generating the list items in a single call.
    single_call_max_tokens : int
        Only used with single_call=True. The maximum number of tokens to generate when generating the list items.
    single_call_top_p : float
        Only used with single_call=True. The top_p to use when generating the list items in a single call.
    
    '''
    block_content = _parser_context["block_content"]
    parser = _parser_context["parser"]
    partial_output = _parser_context["partial_output"]
    parser_prefix = _parser_context["parser_prefix"]
    parser_node = _parser_context["parser_node"]

    assert len(block_content) == 1
    assert not (hidden and single_call), "Cannot use hidden=True and single_call together"
    assert isinstance(list_name, str), "Must provide a variable name to save the generated list to"
    assert not hidden or num_iterations is not None, "Cannot use hidden=True and variable length iteration together yet..."
    echo = not hidden

    # num_iterations has priority over max_iterations if they are both set
    if num_iterations is not None:
        max_iterations = num_iterations
        min_iterations = num_iterations

    # give the list a default name
    if list_name is None:
        list_name = 'generated_list'

    # if stop is None then we use the text of the node after the generate command
    # if stop is None:

    #     next_text = next_node.text if next_node is not None else ""
    #     prev_text = prev_node.text if prev_node is not None else ""

    #     # auto-detect quote stop tokens
    #     quote_types = ['"', "'", "'''", '"""', "`"]
    #     for quote_type in quote_types:
    #         if next_text.startswith(quote_type) and prev_text.endswith(quote_type):
    #             stop = quote_type
    #             break
                
    #     # auto-detect XML tag stop tokens
    #     if stop is None:
    #         m = re.match(r"^\s*(</[^>]+>)", next_text, re.DOTALL) #next_text.startswith(end_tag)
    #         if m is not None:
    #             stop = m.group(1)
            
    #         m = re.match(r"^\s*(<|im_end|>)", next_text, re.DOTALL) #next_text.startswith(end_tag)
    #         if m is not None:
    #             stop = "<|im_end|>"
            
    #         if next_text != "":
    #             stop = next_text

    out = []
    partial_out = ""
    
    # convert stop strings to tokens
    if stop is not False:
        if stop is None:
            stop_tokens = []
        else:
            stop_tokens = [parser.program.llm.encode(s) for s in stop]

    if not single_call:
        i = 0
        data = []
        while True:
            parser.variable_stack.append({})
            parser.variable_stack[-1]["@index"] = i
            parser.variable_stack[-1]["@first"] = i == 0
            parser.variable_stack[-1]["this"] = {}
            pos = len(parser.prefix)
            
            # add the join string if we are not on the first iteration
            if len(data) > 0 and join != "":
                partial_output(join)
            
            await parser.visit(block_content[0]) # fills out parser.prefix
            block_variables = parser.variable_stack.pop()["this"]
            data.append(block_variables)
            if hidden:
                # new_content = parser.prefix[pos:]
                parser.reset_prefix(pos)
            if not parser.executing:
                block_text = parser.prefix[pos:]
                block_text = block_text # make any unfinished this. references point to the last (unfinished) item
                parser.prefix = parser.prefix[:pos] + parser.prefix[pos:].replace("this.", list_name+"[-1].")
                break
            i += 1

            # see if we hit an await and so are not executing anymore
            if not parser.executing:
                break

            # check if the block has thrown a stop iteration signal
            if parser.caught_stop_iteration:
                parser.caught_stop_iteration = False
                break

            # check if we have reached the max iterations
            if max_iterations is not None and i >= max_iterations:
                break

            # we run a quick generation to see if we have reached the end of the list (note the +2 tokens is to help be tolorant to whitespace)
            if stop is not False and i >= min_iterations:
                try:
                    gen_obj = await parser.llm_session(strip_markers(parser.prefix), stop=stop, max_tokens=len(stop_tokens)+2, temperature=0, cache_seed=0)
                except Exception:
                    raise Exception(f"Error generating stop tokens for geneach loop. Perhaps you are outside of role tags (assistant/user/system)? If you don't want the loop to check for stop tokens, set stop=False or set num_iterations.")
                if gen_obj["choices"][0]["finish_reason"] == "stop":
                    break
    
    # TODO: right now single_call is a bit hacky, we should make it more robust to rich loop item template structures
    else: # if single_call
        # create a pattern to match each item
        pattern = re.sub(
            r'{{gen [\'"]([^\'"]+)[\'"][^}]*}}',
            lambda x: r"(?P<"+_escape_group_name(x.group(1))+">.*?)",
            block_content[0].text
        )

        # fixed prefixes can be used if we know we have at least one iteration
        if min_iterations > 0:
            # find what part of the pattern is fixed before the first generation
            fixed_prefix = re.match(r"^(.*)\(\?P\<", pattern, flags=re.DOTALL)[0][:-4]
            fixed_prefix = fixed_prefix.replace(r"{{@index}}", "0") # TODO: this is a bit hacky
        else:
            fixed_prefix = ""

        # assume the LLM will also generate whatever interpolations are in the pattern
        pattern = re.sub(r"{{(.*?)}}", lambda x: r"(?P<" + _escape_group_name(x.group(1)) + ">.*?)", pattern)

        # generate the looped content
        if single_call_temperature > 0:
            cache_seed = parser.program.cache_seed
            parser.program.cache_seed += 1
        else:
            cache_seed = 0
        gen_stream = await parser.llm_session(parser_prefix+fixed_prefix, stop=stop, max_tokens=single_call_max_tokens, temperature=single_call_temperature, top_p=single_call_top_p, cache_seed=cache_seed, stream=True)
        generated_value = fixed_prefix
        num_items = 0
        data = []
        for gen_obj in gen_stream:
            generated_value += gen_obj["choices"][0]["text"]


            # parse the generated content (this assumes the generated content is syntactically correct)
            matches = re.finditer(pattern, generated_value)
            for m in matches:#"{{!--" + f"GMARKER_START_{name}${node_text}$}}{out}{{!--GMARKER_END_{name}$$" + "}}"
                
                # consume the generated value up to the match
                generated_value = generated_value[m.end():]

                # get the variables that were generated
                match_dict = m.groupdict()
                if "this" in match_dict:
                    data.append(match_dict["this"])
                else:
                    d = {}
                    for k in match_dict:
                        k = _unescape_group_name(k)
                        if k.startswith("this."):
                            d[k[5:]] = match_dict[k].strip()
                    data.append(d)

                # recreate the output string with format markers added
                item_out = re.sub(
                    r"{{(?!~?gen)(.*?)}}",
                    lambda x: match_dict[_escape_group_name(x.group(1))],
                    block_content[0].text
                )
                item_out = re.sub(
                    r"{{gen [\'\"]([^\'\"]+)[\'\"][^}]*}}",
                    lambda x: "{{!--GMARKER_START_gen$"+x.group().replace("$", "&#36;").replace("{", "&#123;").replace("}", "&#125;")+"$--}}"+match_dict[_escape_group_name(x.group(1))]+"{{!--GMARKER_END_gen$$--}}",
                    item_out
                )
                partial_output("{{!--GMARKER_each$$--}}" + item_out) # marker and content of the item
                num_items += 1
                # out.append(item_out)

                # if we have hit the max iterations, stop the LLM
                if num_items >= max_iterations:
                    gen_stream.close()
    
    partial_output("{{!--GMARKER_each$$--}}") # end marker

    # parser.get_variable(list, [])
    parser.set_variable(list_name, parser.get_variable(list_name, default_value=[]) + data)
   
    # if we have stopped executing, we need to add the loop to the output so it can be executed later
    if not parser.executing:
        if isinstance(list_name, str):
            partial_output(parser_node.text)

    return ""
    
    if echo:
        return "{{!--GMARKER_each$$--}}" + "{{!--GMARKER_each$$--}}".join(out) + "{{!--GMARKER_each$$--}}" + suffix
    else:
        id = uuid.uuid4().hex
        l = len(out)
        out_str = prefix + "{{!--" + f"GMARKER_each_noecho_start_{echo}_{l}${id}$" + "--}}"
        for i, value in enumerate(out):
            if i > 0:
                out_str += "{{!--" + f"GMARKER_each_noecho_{echo}_{i}${id}$" + "--}}"
            out_str += value
        return out_str + "{{!--" + f"GMARKER_each_noecho_end${id}$" + "--}}"

        # return "{{!--GMARKER_each_noecho$$}}" + "{{!--GMARKER_each_noecho$$}}".join(out) + "{{!--GMARKER_each_noecho$$}}"
geneach.is_block = True

def _escape_group_name(name):
    return name.replace("@", "_AT_").replace(".", "_DOT_")

def _unescape_group_name(name):
    return name.replace("_AT_", "@").replace("_DOT_", ".")